/*
 * Copyright (c) 2002-2022, City of Paris
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice
 *     and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice
 *     and the following disclaimer in the documentation and/or other materials
 *     provided with the distribution.
 *
 *  3. Neither the name of 'Mairie de Paris' nor 'Lutece' nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * License 1.0
 */
package fr.paris.lutece.plugins.appointment.modules.solrsearchapp.service;

import java.time.LocalDateTime;
import java.time.format.DateTimeParseException;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.solr.client.solrj.SolrQuery;
import org.apache.solr.client.solrj.util.ClientUtils;

public class SolrQueryService
{
    private static final String SOLR_FIELD_SITE = "site";
    private static final String SOLR_FIELD_CATEGORY = "categorie";
    public static final String SOLR_FIELD_FORM_UID_TITLE = "form_id_title_string";
    private static final String SOLR_FIELD_FORM_UID = "uid_form_string";
    private static final String SOLR_QUERY_ALL = "*:*";
    private static final String SOLR_FILTERQUERY_ALLOWED_NOW = "{!frange l=0}sub(sub(ms(date),mul(3600000,min_hours_before_appointment_long)),ms())";
    private static final String SOLR_FILTERQUERY_DAY_OPEN = "day_open_string:true";
    private static final String SOLR_FILTERQUERY_ENABLED = "enabled_string:true";
    private static final String SOLR_FILTERQUERY_ACTIVE = "appointment_active_string:true";
    private static final String SOLR_FIELD_TYPE = "type";
    public static final String SOLR_FIELD_DATE = "date";
    private static final String SOLR_FIELD_MINUTE_OF_DAY = "minute_of_day_long";
    private static final String SOLR_NB_CONSECUTIVES_SLOTS = "nb_consecutives_slots_long";
    private static final String SOLR_ROLE = "role";
    public static final String SOLR_FIELD_DAY_OF_WEEK = "day_of_week_long";
    private static final String SOLR_TYPE_APPOINTMENT_SLOT = "appointment-slot";
    public static final String VALUE_FQ_EMPTY = "__EMPTY__";

    public static final List<SimpleImmutableEntry<String, String>> EXACT_FACET_QUERIES = Collections
            .unmodifiableList( Arrays.asList( new SimpleImmutableEntry<>( SOLR_FIELD_SITE, Utilities.PARAMETER_SITE ),
                    new SimpleImmutableEntry<>( SOLR_FIELD_CATEGORY, Utilities.PARAMETER_CATEGORY ),
                    new SimpleImmutableEntry<>( SOLR_FIELD_FORM_UID, Utilities.PARAMETER_FORM ) ) );

    public static final List<SimpleImmutableEntry<String, String>> FACET_FIELDS = Collections
            .unmodifiableList( Arrays.asList( new SimpleImmutableEntry<>( SOLR_FIELD_SITE, Utilities.MARK_ITEM_SITES ),
                    new SimpleImmutableEntry<>( SOLR_FIELD_CATEGORY, Utilities.MARK_ITEM_CATEGORIES ),
                    new SimpleImmutableEntry<>( SOLR_FIELD_FORM_UID_TITLE, Utilities.MARK_ITEM_FORMS ) ) );

    private SolrQueryService( )
    {
        // private constructor
    }

    public static SolrQuery getCommonFilteredQuery( HttpServletRequest request, Map<String, String> searchParameters,
            Map<String, String [ ]> searchMultiParameters )
    {
        SolrQuery query = new SolrQuery( );
        query.setQuery( SOLR_QUERY_ALL );
        query.addFilterQuery( SOLR_FIELD_TYPE + ":" + SOLR_TYPE_APPOINTMENT_SLOT );
        query.addFilterQuery( SOLR_FILTERQUERY_ALLOWED_NOW );
        query.addFilterQuery( SOLR_FILTERQUERY_DAY_OPEN );
        query.addFilterQuery( SOLR_FILTERQUERY_ENABLED );
        query.addFilterQuery( SOLR_FILTERQUERY_ACTIVE );

        for ( SimpleImmutableEntry<String, String> entry : EXACT_FACET_QUERIES )
        {
            addFacetToQuery( query, request, searchParameters, entry );
        }

        StringBuilder sbFqDaysOfWeek = new StringBuilder( );
        String [ ] searchDays = Utilities.getSearchMultiParameter( Utilities.PARAMETER_DAYS_OF_WEEK, request, searchMultiParameters );
        if ( ArrayUtils.isNotEmpty( searchDays ) )
        {
            sbFqDaysOfWeek.append( "{!tag=tag" + SOLR_FIELD_DAY_OF_WEEK + "}" + SOLR_FIELD_DAY_OF_WEEK + ":(" );
            for ( int nDay = 0; nDay < searchDays.length; nDay++ )
            {
                if ( nDay > 0 )
                {
                    sbFqDaysOfWeek.append( " OR " );
                }
                sbFqDaysOfWeek.append( searchDays [nDay] );
            }
            sbFqDaysOfWeek.append( ")" );
        }
        query.addFilterQuery( sbFqDaysOfWeek.toString( ) );
        query.addFacetField( "{!ex=tag" + SOLR_FIELD_DAY_OF_WEEK + "}" + SOLR_FIELD_DAY_OF_WEEK );

        String strFromDate = Utilities.getSearchParameterValue( Utilities.PARAMETER_FROM_DATE, request, searchParameters );
        String strFromTime = Utilities.getSearchParameterValue( Utilities.PARAMETER_FROM_TIME, request, searchParameters );
        String strToDate = Utilities.getSearchParameterValue( Utilities.PARAMETER_TO_DATE, request, searchParameters );
        String strToTime = Utilities.getSearchParameterValue( Utilities.PARAMETER_TO_TIME, request, searchParameters );
        LocalDateTime localDateTimeFrom = null;
        try
        {
            localDateTimeFrom = LocalDateTime.parse( strFromDate + " " + strFromTime, Utilities.inputFormatter );
        }
        catch( DateTimeParseException e )
        {
            localDateTimeFrom = null;
        }
        String strSolrDateTimeFrom = "*";
        if ( localDateTimeFrom != null )
        {
            strSolrDateTimeFrom = localDateTimeFrom.format( Utilities.outputFormatter );
        }
        LocalDateTime localDateTimeTo = null;
        try
        {
            localDateTimeTo = LocalDateTime.parse( strToDate + " " + strToTime, Utilities.inputFormatter );
        }
        catch( DateTimeParseException e )
        {
            localDateTimeTo = null;
        }
        String strSolrDateTimeTo = "*";
        if ( localDateTimeTo != null )
        {
            strSolrDateTimeTo = localDateTimeTo.format( Utilities.outputFormatter );
        }
        query.addFilterQuery( SOLR_FIELD_DATE + ":[" + strSolrDateTimeFrom + " TO " + strSolrDateTimeTo + "]" );
        String strFromDayMinute = Utilities.getSearchParameterValue( Utilities.PARAMETER_FROM_DAY_MINUTE, request, searchParameters );
        String strToDayMinute = Utilities.getSearchParameterValue( Utilities.PARAMETER_TO_DAY_MINUTE, request, searchParameters );
        String strSolrDayMinuteFrom = "*";
        if ( strFromDayMinute != null )
        {
            strSolrDayMinuteFrom = strFromDayMinute;
        }
        String strSolrDayMinuteTo = "*";
        if ( strToDayMinute != null )
        {
            strSolrDayMinuteTo = strToDayMinute;
        }
        query.addFilterQuery( SOLR_FIELD_MINUTE_OF_DAY + ":[" + strSolrDayMinuteFrom + " TO " + strSolrDayMinuteTo + "]" );

        String strNbConsecutiveSlots = Utilities.getSearchParameterValue( Utilities.PARAMETER_NB_SLOTS, request, searchParameters );
        int nbConsecutiveSlots = Integer.parseInt( strNbConsecutiveSlots );
        query.addFilterQuery( SOLR_NB_CONSECUTIVES_SLOTS + ":[" + nbConsecutiveSlots + " TO *]" );

        String strRole = Utilities.getSearchParameterValue( Utilities.PARAMETER_ROLE, request, searchParameters );
        if ( StringUtils.isNotEmpty( strRole ) && !"none".equals( strRole ) )
        {
            query.addFilterQuery( SOLR_ROLE + ":" + strRole );
        }
        return query;
    }

    private static void addFacetToQuery( SolrQuery query, HttpServletRequest request, Map<String, String> searchParameters,
            SimpleImmutableEntry<String, String> entry )
    {
        String strValue = Utilities.getSearchParameterValue( entry.getValue( ), request, searchParameters );
        String strFacetField;
        if ( SOLR_FIELD_FORM_UID.equals( entry.getKey( ) ) )
        {
            strFacetField = SOLR_FIELD_FORM_UID_TITLE;
        }
        else
        {
            strFacetField = entry.getKey( );
        }
        if ( StringUtils.isNotBlank( strValue ) )
        {
            String strFilterQuery;
            if ( VALUE_FQ_EMPTY.equals( strValue ) )
            {
                strFilterQuery = entry.getKey( ) + ":" + "\"\" OR (*:* NOT " + entry.getKey( ) + ":*)";
            }
            else
            {
                strFilterQuery = entry.getKey( ) + ":" + ClientUtils.escapeQueryChars( strValue );
            }
            query.addFilterQuery( "{!tag=tag" + strFacetField + "}" + strFilterQuery );
            strFacetField = "{!ex=tag" + strFacetField + "}" + strFacetField;
        }
        query.addFacetField( strFacetField );
    }
}
